/**
 *  ServingXML
 *  
 *  Copyright (C) 2006  Daniel Parker
 *    daniel.parker@servingxml.com 
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License. 
 * 
 **/

package com.servingxml.components.flatfile.options;

import java.nio.charset.Charset;

import com.servingxml.components.flatfile.options.QuoteSymbol;
import com.servingxml.util.Alignment;
import com.servingxml.util.ServingXmlException;
import com.servingxml.components.flatfile.parsing.Token;

/**
 *
 * 
 * @author Daniel A. Parker (daniel.parker@servingxml.com)
 */

public class FlatFileOptionsImpl implements FlatFileOptions {

  private final Charset charset;
  private boolean ignoreTrailingEmptyLines;
  private boolean ignoreEmptyLines;
  private boolean lineDelimited;
  private CommentStarter commentStarter;
  private Delimiter[] recordDelimiters;
  private Delimiter[] segmentDelimiters;
  private Delimiter[] fieldDelimiters;
  private Delimiter[] nameDelimiters;
  private Delimiter[] repeatDelimiters;
  private Delimiter[] subfieldDelimiters;
  private boolean countPositionsInBytes;
  private boolean flushRecordOnWrite;
  private boolean trimLeading;
  private boolean trimTrailing;
  private boolean trimLeadingWithinQuotes;
  private boolean trimTrailingWithinQuotes;
  private QuoteSymbol quoteSymbol;
  private boolean quote;
  private boolean alwaysQuote;
  private boolean omitFinalFieldDelimiter;
  private boolean omitFinalRepeatDelimiter;
  private char padCharacter;
  private Alignment alignment;
  private QuoteSymbolByteChecker byteQuoteSymbolChecker;
  private QuoteSymbolCharChecker charQuoteSymbolChecker;
  private ByteTrimmer byteTrimmer;
  private CharTrimmer charTrimmer;

  private DelimiterByteChecker[] recordByteDelimiterCheckers;
  private DelimiterByteChecker[] segmentByteDelimiterCheckers;
  private DelimiterByteChecker[] repeatByteDelimiterCheckers;
  private DelimiterByteChecker[] tagByteDelimiterCheckers;
  private DelimiterByteChecker[] fieldByteDelimiterCheckers;
  private DelimiterByteChecker[] subfieldByteDelimiterCheckers;

  private DelimiterCharChecker[] recordCharDelimiterCheckers;
  private DelimiterCharChecker[] segmentCharDelimiterCheckers;
  private DelimiterCharChecker[] repeatCharDelimiterCheckers;
  private DelimiterCharChecker[] tagCharDelimiterCheckers;
  private DelimiterCharChecker[] fieldCharDelimiterCheckers;
  private DelimiterCharChecker[] subfieldCharDelimiterCheckers;

  private CommentStarterByteChecker byteCommentStarterChecker;
  private CommentStarterCharChecker charCommentStarterChecker;

  public FlatFileOptionsImpl(Charset charset, boolean delimitedFile, boolean countPositionsInBytes) { 
    this.charset = charset;

    this.ignoreTrailingEmptyLines = true;
    this.ignoreEmptyLines = false;
    this.lineDelimited = true;
    this.flushRecordOnWrite = false;
    this.trimLeading = true;
    this.trimTrailing = true;
    this.trimLeadingWithinQuotes = false;
    this.trimTrailingWithinQuotes = false;
    this.countPositionsInBytes = countPositionsInBytes;
    this.omitFinalFieldDelimiter = true;
    this.omitFinalRepeatDelimiter = true;

    this.commentStarter = CommentStarter.NULL;
    this.byteCommentStarterChecker = CommentStarterByteChecker.NULL;
    this.charCommentStarterChecker = CommentStarterCharChecker.NULL;

    //RecordDelimiter CRLF = new RecordDelimiter(RecordDelimiter.CRLF, true, true);
    //RecordDelimiter LF = new RecordDelimiter(RecordDelimiter.LF, true, true);
    //this.recordDelimiters = new RecordDelimiter[]{CRLF,LF};

    this.recordDelimiters = new RecordDelimiter[]{RecordDelimiter.CRLF,RecordDelimiter.LF};
    this.segmentDelimiters = Delimiter.EMPTY_DELIMITER_ARRAY;
    this.fieldDelimiters = Delimiter.EMPTY_DELIMITER_ARRAY;
    this.nameDelimiters = Delimiter.EMPTY_DELIMITER_ARRAY;
    this.repeatDelimiters = Delimiter.EMPTY_DELIMITER_ARRAY;
    this.subfieldDelimiters = Delimiter.EMPTY_DELIMITER_ARRAY;

    this.recordByteDelimiterCheckers = makeByteDelimiterCheckers(recordDelimiters,charset); 
    this.segmentByteDelimiterCheckers = DelimiterByteChecker.EMPTY_DELIMITER_CHECKER_ARRAY;
    this.repeatByteDelimiterCheckers = DelimiterByteChecker.EMPTY_DELIMITER_CHECKER_ARRAY; 
    this.tagByteDelimiterCheckers = DelimiterByteChecker.EMPTY_DELIMITER_CHECKER_ARRAY;    
    this.fieldByteDelimiterCheckers = DelimiterByteChecker.EMPTY_DELIMITER_CHECKER_ARRAY;
    this.subfieldByteDelimiterCheckers = DelimiterByteChecker.EMPTY_DELIMITER_CHECKER_ARRAY;

    this.recordCharDelimiterCheckers = makeCharDelimiterCheckers(recordDelimiters); 
    this.segmentCharDelimiterCheckers = DelimiterCharChecker.EMPTY_DELIMITER_CHECKER_ARRAY;
    this.repeatCharDelimiterCheckers = DelimiterCharChecker.EMPTY_DELIMITER_CHECKER_ARRAY; 
    this.tagCharDelimiterCheckers = DelimiterCharChecker.EMPTY_DELIMITER_CHECKER_ARRAY;    
    this.fieldCharDelimiterCheckers = DelimiterCharChecker.EMPTY_DELIMITER_CHECKER_ARRAY;
    this.subfieldCharDelimiterCheckers = DelimiterCharChecker.EMPTY_DELIMITER_CHECKER_ARRAY;

    this.quoteSymbol = new QuoteSymbol('"', "\"\"");
    if (delimitedFile) {
      this.quote = true;
    } else {
      this.quote = false;
    }
    this.alwaysQuote = false;
    this.padCharacter = ' ';
    this.alignment = Alignment.LEFT;
    this.byteQuoteSymbolChecker = quoteSymbol.createQuoteSymbolChecker(charset);
    this.charQuoteSymbolChecker = quoteSymbol.createCharQuoteSymbolChecker();
    this.byteTrimmer = ByteTrimmer.newInstance(charset);
    this.charTrimmer = CharTrimmer.newInstance();
  }

  public FlatFileOptionsImpl(FlatFileOptions defaults) { 
    this.ignoreTrailingEmptyLines = defaults.isIgnoreTrailingEmptyLines();
    this.ignoreEmptyLines = defaults.isIgnoreEmptyLines();
    this.lineDelimited = defaults.isLineDelimited();
    this.commentStarter = defaults.getCommentStarter();
    this.recordDelimiters = defaults.getRecordDelimiters();
    this.segmentDelimiters = defaults.getSegmentDelimiters();
    this.fieldDelimiters = defaults.getFieldDelimiters();
    this.nameDelimiters = defaults.getNameDelimiters();
    this.repeatDelimiters = defaults.getRepeatDelimiters();
    this.subfieldDelimiters = defaults.getSubfieldDelimiters();
    this.countPositionsInBytes = defaults.isCountPositionsInBytes();
    this.flushRecordOnWrite = defaults.isFlushRecordOnWrite();
    this.trimLeading = defaults.isTrimLeading();
    this.trimTrailing = defaults.isTrimTrailing();
    this.trimLeadingWithinQuotes = defaults.isTrimLeadingWithinQuotes();
    this.trimTrailingWithinQuotes = defaults.isTrimTrailingWithinQuotes();
    this.quoteSymbol = defaults.getQuoteSymbol();
    this.quote = defaults.isQuote();
    this.alwaysQuote = defaults.isAlwaysQuote();
    this.omitFinalFieldDelimiter = defaults.isOmitFinalFieldDelimiter();
    this.omitFinalRepeatDelimiter = defaults.isOmitFinalRepeatDelimiter();
    this.padCharacter = defaults.getPadCharacter();
    this.alignment = defaults.getAlignment();
    this.charset = defaults.getCharset();
    this.byteQuoteSymbolChecker = defaults.getQuoteSymbolByteChecker();
    this.charQuoteSymbolChecker = defaults.getQuoteSymbolCharChecker();
    this.byteTrimmer = defaults.getByteTrimmer();
    this.charTrimmer = defaults.getCharTrimmer();

    this.recordByteDelimiterCheckers = defaults.getRecordDelimiterByteCheckers(); 
    this.segmentByteDelimiterCheckers = defaults.getSegmentDelimiterByteCheckers();
    this.repeatByteDelimiterCheckers = defaults.getRepeatDelimiterByteCheckers(); 
    this.tagByteDelimiterCheckers = defaults.getNameDelimiterByteCheckers();    
    this.fieldByteDelimiterCheckers = defaults.getFieldDelimiterByteCheckers();
    this.subfieldByteDelimiterCheckers = defaults.getSubfieldDelimiterByteCheckers();

    this.recordCharDelimiterCheckers = defaults.getRecordDelimiterCharCheckers(); 
    this.segmentCharDelimiterCheckers = defaults.getSegmentDelimiterCharCheckers();
    this.repeatCharDelimiterCheckers = defaults.getRepeatDelimiterCharCheckers(); 
    this.tagCharDelimiterCheckers = defaults.getNameDelimiterCharCheckers();    
    this.fieldCharDelimiterCheckers = defaults.getFieldDelimiterCharCheckers();
    this.subfieldCharDelimiterCheckers = defaults.getSubfieldDelimiterCharCheckers();

    this.byteCommentStarterChecker = defaults.getCommentStarterByteChecker();
    this.charCommentStarterChecker = defaults.getCommentStarterCharChecker();
  }

  public Charset getDefaultCharset() {
    return charset == null ? Charset.defaultCharset() : charset;
  }

  public Charset getCharset() {
    return charset;
  }

  public ByteTrimmer getByteTrimmer() {
    return byteTrimmer;
  }

  public CharTrimmer getCharTrimmer() {
    return charTrimmer;
  }

  public QuoteSymbolByteChecker getQuoteSymbolByteChecker() {
    return isQuote() ? byteQuoteSymbolChecker : QuoteSymbolByteChecker.NULL; 
  }

  public QuoteSymbolCharChecker getQuoteSymbolCharChecker() {
    return isQuote() ? charQuoteSymbolChecker : QuoteSymbolCharChecker.NULL; 
  }

  public QuoteSymbol getQuoteSymbol() {
    return quoteSymbol;
  }

  public boolean isQuote() {
    return quote;
  }

  public void setQuote(boolean quote) {
    this.quote = quote;
  }

  public boolean isAlwaysQuote() {
    return alwaysQuote;
  }

  public void setAlwaysQuote(boolean alwaysQuote) {
    this.alwaysQuote = alwaysQuote;
  }

  public void setQuoteSymbol(QuoteSymbol quoteSymbol) {
    this.quoteSymbol = quoteSymbol;
    this.byteQuoteSymbolChecker = quoteSymbol.createQuoteSymbolChecker(charset);
    this.charQuoteSymbolChecker = quoteSymbol.createCharQuoteSymbolChecker();
  }

  public boolean isIgnoreTrailingEmptyLines() {
    return ignoreTrailingEmptyLines;
  }

  public void setIgnoreTrailingEmptyLines(boolean ignoreTrailingEmptyLines) {
    this.ignoreTrailingEmptyLines = ignoreTrailingEmptyLines;
  }

  public boolean isIgnoreEmptyLines() {
    return ignoreEmptyLines;
  }

  public void setIgnoreEmptyLines(boolean ignoreEmptyLines) {
    this.ignoreEmptyLines = ignoreEmptyLines;
  }

  public boolean isLineDelimited() {
    return lineDelimited;
  }

  public void setLineDelimited(boolean lineDelimited) {
    this.lineDelimited = lineDelimited;
  }
  public boolean isOmitFinalFieldDelimiter() {
    return omitFinalFieldDelimiter;
  }

  public boolean isOmitFinalRepeatDelimiter() {
    return omitFinalRepeatDelimiter;
  }

  public void setOmitFinalRepeatDelimiter(boolean omitFinalRepeatDelimiter) {
    this.omitFinalRepeatDelimiter = omitFinalRepeatDelimiter;
  }

  public CommentStarter getCommentStarter() {
    return commentStarter;
  }

  public CommentStarterByteChecker getCommentStarterByteChecker() {
    return byteCommentStarterChecker;
  }

  public CommentStarterCharChecker getCommentStarterCharChecker() {
    return charCommentStarterChecker;
  }

  public void setCommentStarter(CommentStarter commentStarter) {
    this.commentStarter = commentStarter;
    this.byteCommentStarterChecker = commentStarter.createByteCommentStarterChecker(charset);
    this.charCommentStarterChecker = commentStarter.createCharCommentStarterChecker();
  }

  public Delimiter getRecordDelimiterForWriting() {
    int index = -1;
    boolean found = false;
    String lineSep = System.getProperty("line.separator");
    for (int i = recordDelimiters.length-1; !found && i >= 0; --i) {
      Delimiter delimiter = recordDelimiters[i];
      if (delimiter.forWriting()) {
        index = i;
        if (delimiter.equalsString(lineSep)) {
          //System.out.println(getClass().getName()+".cons system delimiter found" );
          found = true;
        }
      }
    }
    if (index == -1) {
      throw new ServingXmlException("Cannot find record delimiter for writing.");
    }
    Delimiter recordDelimiter = recordDelimiters[index];
    return recordDelimiter;
  }


  public void setOmitFinalFieldDelimiter(boolean omitFinalFieldDelimiter) {
    this.omitFinalFieldDelimiter = omitFinalFieldDelimiter;
  }

  public Delimiter[] getRecordDelimiters() {
    return recordDelimiters;
  }

  public void setRecordDelimiters(Delimiter[] recordDelimiters) {
    //System.out.println(getClass().getName()+".setRecordDelimiters");
    //for (int i = 0; i < recordDelimiters.length; ++i) {
      //System.out.println(recordDelimiters[i].getClass().getName());
    //}
    this.recordDelimiters = recordDelimiters;
    this.recordByteDelimiterCheckers = makeByteDelimiterCheckers(recordDelimiters, charset);
    this.recordCharDelimiterCheckers = makeCharDelimiterCheckers(recordDelimiters);
  }

  public Delimiter[] getSegmentDelimiters() {
    return segmentDelimiters;
  }

  public void setSegmentDelimiters(Delimiter[] segmentDelimiters) {
    this.segmentDelimiters = segmentDelimiters;
    this.segmentByteDelimiterCheckers = makeByteDelimiterCheckers(segmentDelimiters, charset);
    this.segmentCharDelimiterCheckers = makeCharDelimiterCheckers(segmentDelimiters);
  }

  public Delimiter[] getFieldDelimiters() {
    return fieldDelimiters;
  }

  public void setFieldDelimiters(Delimiter[] fieldDelimiters) {
    this.fieldDelimiters = fieldDelimiters;
    this.fieldByteDelimiterCheckers = makeByteDelimiterCheckers(fieldDelimiters, charset);
    this.fieldCharDelimiterCheckers = makeCharDelimiterCheckers(fieldDelimiters);
  }

  public Delimiter[] getNameDelimiters() {
    return nameDelimiters;
  }

  public void setNameDelimiters(Delimiter[] nameDelimiters) {
    this.nameDelimiters = nameDelimiters;
    this.tagByteDelimiterCheckers = makeByteDelimiterCheckers(nameDelimiters, charset);
    this.tagCharDelimiterCheckers = makeCharDelimiterCheckers(nameDelimiters);
  }

  public Delimiter[] getRepeatDelimiters() {
    return repeatDelimiters;
  }

  public void setRepeatDelimiters(Delimiter[] repeatDelimiters) {
    this.repeatDelimiters = repeatDelimiters;
    this.repeatByteDelimiterCheckers = makeByteDelimiterCheckers(repeatDelimiters, charset);
    this.repeatCharDelimiterCheckers = makeCharDelimiterCheckers(repeatDelimiters);
  }

  public Delimiter[] getSubfieldDelimiters() {
    return subfieldDelimiters;
  }

  public void setSubfieldDelimiters(Delimiter[] subfieldDelimiters) {
    this.subfieldDelimiters = subfieldDelimiters;
    this.subfieldByteDelimiterCheckers = makeByteDelimiterCheckers(subfieldDelimiters, charset);
    this.subfieldCharDelimiterCheckers = makeCharDelimiterCheckers(subfieldDelimiters);
  }

  public boolean isCountPositionsInBytes() {
    //System.out.println(getClass().getName()+".isCountPositionsInBytes countPositionsInBytes="+countPositionsInBytes);
    return countPositionsInBytes;
  }

  public void setCountPositionsInBytes(boolean countPositionsInBytes) {
    //System.out.println(getClass().getName()+".setCountPositionsInBytes countPositionsInBytes="+countPositionsInBytes);
    this.countPositionsInBytes = countPositionsInBytes;
  }

  public boolean isTrimLeading() {
    return trimLeading;
  }

  public boolean isTrimTrailing() {
    return trimTrailing;
  }

  public boolean isTrimLeadingWithinQuotes() {
    return trimLeadingWithinQuotes;
  }

  public boolean isTrimTrailingWithinQuotes() {
    return trimTrailingWithinQuotes;
  }

  public void setTrimLeading(boolean trimLeading) {
    this.trimLeading = trimLeading;
  }

  public void setTrimTrailing(boolean trimTrailing) {
    this.trimTrailing = trimTrailing;
  }

  public void setTrimLeadingWithinQuotes(boolean trimLeadingWithinQuotes) {
    //System.out.println(getClass().getName()+".setTrimLeadingWithinQuotes trimLeadingWithinQuotes="+trimLeadingWithinQuotes);

    this.trimLeadingWithinQuotes = trimLeadingWithinQuotes;
  }

  public void setTrimTrailingWithinQuotes(boolean trimTrailingWithinQuotes) {
    //System.out.println(getClass().getName()+".setTrimTrailingWithinQuotes trimTrailingWithinQuotes="+trimTrailingWithinQuotes);
    this.trimTrailingWithinQuotes = trimTrailingWithinQuotes;
  }

  public boolean isFlushRecordOnWrite() {
    return flushRecordOnWrite;
  }

  public void setFlushRecordOnWrite(boolean flushRecordOnWrite) {
    this.flushRecordOnWrite = flushRecordOnWrite;
  }

  public Alignment getAlignment() {
    return alignment;
  }

  public void setAlignment(Alignment alignment) {
    this.alignment = alignment;
  }

  public char getPadCharacter() {
    return padCharacter;
  }

  public void setPadCharacter(char padCharacter) {
    this.padCharacter = padCharacter;
  }

  public boolean useQuotes(String value) {
    //System.out.println(getClass().getName()+".useQuotes field delim count = " + fieldDelimiters.length);
    boolean use = alwaysQuote;
    if (!alwaysQuote && quote) {
      for (int i = 0; !use && i < recordDelimiters.length; ++i) {
        use = recordDelimiters[i].occursIn(value);
      }
      for (int i = 0; !use && i < segmentDelimiters.length; ++i) {
        use = segmentDelimiters[i].occursIn(value);
      }
      for (int i = 0; !use && i < fieldDelimiters.length; ++i) {
        //StringBuilder buf=new StringBuilder();
        //fieldDelimiters[i].writeEndDelimiterTo(buf);
        //System.out.println(getClass().getName()+".useQuotes field delim = " + buf);
        use = fieldDelimiters[i].occursIn(value);
      }
      //for (int i = 0; !use && i < nameDelimiters.length; ++i) {
      //  use = nameDelimiters[i].occursIn(value);
      //}
      for (int i = 0; !use && i < repeatDelimiters.length; ++i) {
        use = repeatDelimiters[i].occursIn(value);
      }
      for (int i = 0; !use && i < subfieldDelimiters.length; ++i) {
        use = subfieldDelimiters[i].occursIn(value);
      }
    }
    //System.out.println(getClass().getName()+".useQuotes |" + value + "| quote="+quote+", always=" + alwaysQuote + ", use=" + use);

    return use;
  }

  public DelimiterByteChecker[] getRecordDelimiterByteCheckers() {
    return recordByteDelimiterCheckers;
  }

  public DelimiterByteChecker[] getSegmentDelimiterByteCheckers() {
    return segmentByteDelimiterCheckers;
  }

  public DelimiterByteChecker[] getRepeatDelimiterByteCheckers() {
    return repeatByteDelimiterCheckers;
  }

  public DelimiterByteChecker[] getNameDelimiterByteCheckers() {
    return tagByteDelimiterCheckers;
  }

  public DelimiterByteChecker[] getFieldDelimiterByteCheckers() {
    return fieldByteDelimiterCheckers;
  }

  public DelimiterByteChecker[] getSubfieldDelimiterByteCheckers() {
    return subfieldByteDelimiterCheckers;
  }

  public DelimiterCharChecker[] getRecordDelimiterCharCheckers() {
    return recordCharDelimiterCheckers;
  }

  public DelimiterCharChecker[] getSegmentDelimiterCharCheckers() {
    return segmentCharDelimiterCheckers;
  }

  public DelimiterCharChecker[] getRepeatDelimiterCharCheckers() {
    return repeatCharDelimiterCheckers;
  }

  public DelimiterCharChecker[] getNameDelimiterCharCheckers() {
    return tagCharDelimiterCheckers;
  }

  public DelimiterCharChecker[] getFieldDelimiterCharCheckers() {
    return fieldCharDelimiterCheckers;
  }

  public DelimiterCharChecker[] getSubfieldDelimiterCharCheckers() {
    return subfieldCharDelimiterCheckers;
  }

  private DelimiterByteChecker[] makeByteDelimiterCheckers(Delimiter[] delimiters, Charset charset) {
    DelimiterByteChecker[] delimiterCheckers;
    delimiterCheckers = new DelimiterByteChecker[delimiters.length];
    for (int i = 0; i < delimiters.length; ++i) {
      delimiterCheckers[i] = delimiters[i].createDelimiterChecker(charset);
    }
    return delimiterCheckers;
  }

  private DelimiterCharChecker[] makeCharDelimiterCheckers(Delimiter[] delimiters) {
    DelimiterCharChecker[] delimiterCheckers;
    delimiterCheckers = new DelimiterCharChecker[delimiters.length];
    for (int i = 0; i < delimiters.length; ++i) {
      delimiterCheckers[i] = delimiters[i].createCharDelimiterChecker();
    }
    return delimiterCheckers;
  }
}

